# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This package contains analytic functions useful for astronomy.

In future versions of ``astropy``, many of these might be
accessible as `~astropy.modeling.core.Model`.

"""

# Shortcuts for most commonly used blackbody functions
from .blackbody import blackbody_nu, blackbody_lambda
