# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module contains configuration and setup utilities for the
Astropy project. This includes all functionality related to the
affiliated package index.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from .paths import *
from .configuration import *
from .affiliated import *
